/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.util.io;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.lang.reflect.Field;
import java.nio.ByteBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.ReadableByteChannel;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL
 * Author : Peter Nedonosko
 *          peter.nedonosko@exoplatform.com.ua
 * 12.09.2006
 * @author <a href="mailto:peter.nedonosko@exoplatform.com.ua">Peter Nedonosko</a>
 * @version $Id: BLOBUtil.java 17644 2007-07-10 21:22:20Z peterit $
 */
public class BLOBUtil {

  private static Log log = ExoLogger.getLogger("jcr.BLOBUtil");
  
  public static final int BLOB_MIN_AVAILABLE = 2048; // 2Kb default BLOB stream available value
  
  public static final int BLOB_MIN_INMEMORY = 128; // 128b BLOB size stored in memory
  
  public static final int BLOB_MIN_SIZE = 2 * 1024 * 1024; // 2Mb default BLOB file size
  
  public static final int BLOB_BUFFER_SIZE = 200 * 1024; // 200Kb default binary values size
  public static final int BLOB_BUFFER_SIZE2 = 2 * 1024 * 1024; // 2Mb big files binary values size
  public static final int BLOB_BUFFER_SIZE3 = 20 * 1024 * 1024; // 20Mb very big files binary values size
  
  /**
   * Calc an optimal memory buffer size for BLOB read opeartions
   * 
   * @param availableBytes, bytes cout available for reading (e.g. Stream.available() result)  
   * 
   * @return allocation buffer size
   */
  static public int calcBufferSize(int availableBytes) {
    
    final long maval = Runtime.getRuntime().maxMemory();
    // if half of available memory great than of max buffer size use buff size #3,
    // otherwise use buff size #2
    final int maxBufferSize = maval >>> 1 > BLOB_BUFFER_SIZE3 ? BLOB_BUFFER_SIZE3 : BLOB_BUFFER_SIZE2;
    
    return        availableBytes > BLOB_MIN_INMEMORY ? 
                       availableBytes > BLOB_MIN_AVAILABLE ? // 2K 
                           availableBytes > BLOB_BUFFER_SIZE ?  // 200K
                               availableBytes > BLOB_BUFFER_SIZE2 ? // 2M
                                   availableBytes > maxBufferSize ? maxBufferSize : BLOB_BUFFER_SIZE2 
                               : BLOB_BUFFER_SIZE >>> 1 // 100K 
                           : availableBytes   
                       : BLOB_MIN_AVAILABLE >>> 1 // 1K
                   : BLOB_MIN_INMEMORY;
  }
  
  static public byte[] allocateBuffer(int availableBytes) {
    
    return new byte[calcBufferSize(availableBytes)];
  }
  
  static public byte[] wrapByteArrayInputStream(final ByteArrayInputStream aStream) {
    //  reflection magic...
    boolean unsetAccessible = false;
    Field buff = null;
    try {
      buff = aStream.getClass().getDeclaredField("buf");
      if (!buff.isAccessible())
        try {
          buff.setAccessible(true);
          unsetAccessible = true;
        } catch (SecurityException e) {
          log.warn("Can't change field 'buf' accessible status in object of the ByteArrayInputStream class", e);
        }
        return (byte[]) buff.get(aStream);
    } catch (NoSuchFieldException e) {
      if (log.isDebugEnabled())
        log.debug("Stream " + aStream + " doesn't contains field with name 'buf'", e);
    } catch (IllegalArgumentException e) {
      if (log.isDebugEnabled())
        log.debug("Check that specified object is an instance of the ByteArrayInputStream class", e);
    } catch (IllegalAccessException e) {
      if (log.isDebugEnabled())
        log.debug("Check that field 'buf' is accessible in object of the ByteArrayInputStream class", e);
    } finally {
      if (unsetAccessible && buff != null)
        buff.setAccessible(false);
    }
    return null;
  }
  
  public static long calcStreamLength(final InputStream aStream) throws IOException {
    
    // some magic... 
    if (aStream instanceof BytesArray) {
      
      return ((BytesArray) aStream).size();
    } else if (aStream instanceof FileInputStream) {
      
      final FileChannel channel = ((FileInputStream) aStream).getChannel();
      if (channel != null)
        return channel.size();
    } else if (aStream instanceof LazyFileInputStream) {
      
      final FileChannel channel = ((LazyFileInputStream) aStream).getChannel();
      if (channel != null)
        return channel.size();
    } else if (aStream instanceof ByteArrayInputStream) {
      ByteArrayInputStream bais = (ByteArrayInputStream) aStream;
      final byte[] buffValue = wrapByteArrayInputStream(bais);
      if (buffValue != null)
        return buffValue.length;
    }
    
    return aStream.available();
  }
  
  /**
   * Read the stream in-memory bytes array if the stream size 
   * is not more BLOB_MIN_INMEMORY size only.
   * 
   * @param aStream
   * @return
   * @throws IOException
   */
  public static byte[] readStreamInmemory(final InputStream aStream) {
    try {
      if (aStream instanceof BytesArray) {
        
        byte[] bytes = ((BytesArray) aStream).getBytes();
        if (bytes.length <= BLOB_MIN_INMEMORY)
          return copyBytes(bytes); 
        
      } else if (aStream instanceof ByteArrayInputStream) {
        
        ByteArrayInputStream bais = (ByteArrayInputStream) aStream;
        byte[] bytes = wrapByteArrayInputStream(bais);
        if (bytes.length <= BLOB_MIN_INMEMORY)
          return copyBytes(bytes);
        
      } else {
      
        byte[] buff = new byte[BLOB_MIN_INMEMORY];
     
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        int res = 0;
        while ((res = aStream.read(buff)) > 0) {
          if (baos.size() > BLOB_MIN_INMEMORY) {
            baos = null;
            break;
          }
          baos.write(buff, 0, res);
        }
        
        if (baos != null)
          return baos.toByteArray();
      }
    } catch(IOException e) {
      throw new IllegalStateException("Can't read stream data in in-memory bytes array: " + e.getMessage(), e);
    }
    
    throw new IllegalStateException("Data could not be rendered as byte array (large data)");
  }
    
  public static byte[] readValue(ValueData data) throws IOException { 
    return data.isByteArray() ? data.getAsByteArray() : readStream(data.getAsStream());
  }
  
  public static byte[] copyBytes(final byte[] source) {
    byte[] copyBuff = new byte[source.length];  
    System.arraycopy(source, 0, copyBuff, 0, copyBuff.length);
    return copyBuff;
  }
  
  /**
   * Optimized for performance InputStream read method
   * 
   * @param aStream
   * @return a stream content in bytes array
   * @throws IOException
   */
  public static byte[] readStream(final InputStream aStream) throws IOException {
    if (aStream instanceof FileInputStream) {
      return readFile((FileInputStream) aStream);
    } else if (aStream instanceof BytesArray) {
      return copyBytes(((BytesArray) aStream).getBytes());
    } else {
      if (aStream instanceof ByteArrayInputStream) {
        byte[] bytes = wrapByteArrayInputStream((ByteArrayInputStream) aStream);
        if (bytes != null) {
          return copyBytes(bytes);
        }
      }
      return readInputStream(aStream);
    }
  }  
  
  /**
   * WARNING - Check that you have enough memory in case of BLOBs.
   */
  static public byte[] readFile(final FileInputStream file) throws IOException {
    
    final FileChannel channel = file.getChannel();
    try {
      final int size = (int) channel.size();
      
      if (size <= BLOB_MIN_SIZE) {
        // reading in classical way
        return readInputStream(file); 
      }
      
      final ByteBuffer fileBuffer = ByteBuffer.allocate(size); 
      
      channel.read(fileBuffer);
      fileBuffer.rewind(); // [PN] 24.05.2007 BUGFIX. Rewind the buffer to begin
      final byte[] buff = new byte[size];
      
      fileBuffer.get(buff);
      fileBuffer.clear();
      
      return buff;
    } finally {
      if (channel != null)
        channel.close();
    }
  }  
  
  /**
   * WARNING: RandomAccessFile in case of BLOB files uses lot of available memory.
   * 
   * @param source
   * @param destFile
   * @return
   * @throws IOException
   */
  @Deprecated
  public static long copyFile(FileInputStream source, RandomAccessFile destFile) throws IOException {
    FileChannel sourceChannel = null;
    FileChannel destenationChannel = null;
    try {
      
      sourceChannel = source.getChannel();
      destenationChannel = destFile.getChannel();
      
      sourceChannel.transferTo(0, sourceChannel.size(), destenationChannel);
      
      return destenationChannel.size();
    } finally {
      if (destenationChannel != null)
        destenationChannel.close();
      if (sourceChannel != null)
        sourceChannel.close();
    }
  }
  
  public static long copyFile(final FileInputStream source, final FileOutputStream destenation) throws IOException {
    
    final FileChannel sourceChannel = source.getChannel(); 
    final FileChannel destenationChannel = destenation.getChannel();
    try {
      sourceChannel.transferTo(0, sourceChannel.size(), destenationChannel);      
      return sourceChannel.size();
    } finally {
      destenationChannel.force(true);
      destenationChannel.close();
      sourceChannel.close();
    }
  }
  
  /**
   * WARNING: RandomAccessFile in case of BLOB files uses lot of available memory.
   * 
   * @param buff - source
   * @param outFile - dstenation
   * @return
   * @throws IOException
   */
  @Deprecated
  public static long writeFile(byte[] buff, FileOutputStream fileOutput) throws IOException {

    FileChannel channel = fileOutput.getChannel();
    MappedByteBuffer storageBuffer = channel.map(
        FileChannel.MapMode.READ_WRITE, 
        0,  
        buff.length);
    storageBuffer.put(buff, 0, buff.length);
    storageBuffer.force();
    storageBuffer.clear();
        
    long size = channel.size();
    
    channel.close();
    
    return size;
  }
  
  /**
   * @param inputStream - source
   * @param outFile - destenation
   * @return - bytes count
   * @throws IOException
   */
  static public long writeFile(final InputStream inputStream, final FileOutputStream outFile) throws IOException {
    
    if (inputStream instanceof FileInputStream) {
      return BLOBUtil.copyFile((FileInputStream) inputStream, outFile);
    } else if (inputStream instanceof BytesArray) {
      byte[] buff = ((BytesArray) inputStream).getBytes();
      outFile.write(buff);
      return buff.length;
    } else if (inputStream instanceof LazyFileInputStream) {
      final FileChannel sourceChannel = ((LazyFileInputStream) inputStream).getChannel();
      final FileChannel destenationChannel = outFile.getChannel();
      try {
        sourceChannel.transferTo(0, sourceChannel.size(), destenationChannel);
        
        return destenationChannel.size();
        
      } finally {
        destenationChannel.force(true);
        destenationChannel.close();
        sourceChannel.close();
      }  
    } else {
      // Try as bytes if ByteArrayInputStream
      if (inputStream instanceof ByteArrayInputStream) {
        byte[] buffValue = wrapByteArrayInputStream((ByteArrayInputStream) inputStream);
        if (buffValue != null) {
          outFile.write(buffValue);
          return buffValue.length;
        }
      }    
      
      final ReadableByteChannel input = Channels.newChannel(inputStream);
      final FileChannel output = outFile.getChannel();
      
      final int available = inputStream.available() + 64;
      final int buffSize = BLOBUtil.calcBufferSize(available);
      
      final ByteBuffer storageBuffer = available >= BLOB_MIN_AVAILABLE ? ByteBuffer.allocateDirect(buffSize) : ByteBuffer.allocate(buffSize);
      
      // [PN] check if we need sync read cycle in concurrent environment
      int res = 0;
      long size = 0;
      while ((res = input.read(storageBuffer)) >= 0 || (res < 0 && !storageBuffer.hasRemaining())) {  
        // Set the limit to the current position and the position to 0
        // making the new bytes visible for write()
        storageBuffer.flip();

        // Write the bytes to the channel
        int numWritten = output.write(storageBuffer);

        // Check if all bytes were written
        if (storageBuffer.hasRemaining()) {
            // If not all bytes were written, move the unwritten bytes
            // to the beginning and set position just after the last
            // unwritten byte; also set limit to the capacity
          storageBuffer.compact();
        } else {
            // Set the position to 0 and the limit to capacity
          storageBuffer.clear();
        }
        
        size += numWritten;
      }
      
      output.close();
      input.close();
      
      storageBuffer.clear();
        
      return size;  
    }
  }

  static protected byte[] readInputStream(final InputStream aStream) throws IOException {
    // optimizing small size data reading and memory usage till stream read operation
    byte[] buff = allocateBuffer(aStream.available());

    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    int res = 0;
    while ((res = aStream.read(buff)) > 0) {
      baos.write(buff, 0, res);
    }
    
    return baos.toByteArray();
  }
}
