/**
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core.nodetype;

import java.util.ArrayList;
import java.util.List;

import javax.jcr.InvalidItemStateException;
import javax.jcr.ItemExistsException;
import javax.jcr.PathNotFoundException;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;
import javax.jcr.nodetype.ConstraintViolationException;
import javax.jcr.nodetype.NodeDefinition;
import javax.jcr.nodetype.NodeType;
import javax.jcr.nodetype.PropertyDefinition;
import javax.jcr.version.OnParentVersionAction;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.AccessControlEntry;
import org.exoplatform.services.jcr.access.AccessControlList;
import org.exoplatform.services.jcr.core.ExtendedPropertyType;
import org.exoplatform.services.jcr.dataflow.DataManager;
import org.exoplatform.services.jcr.dataflow.ItemState;
import org.exoplatform.services.jcr.datamodel.IllegalNameException;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.LocationFactory;
import org.exoplatform.services.jcr.impl.core.value.NameValue;
import org.exoplatform.services.jcr.impl.core.value.ValueFactoryImpl;
import org.exoplatform.services.jcr.impl.dataflow.MutableItemDataChangesLog;
import org.exoplatform.services.jcr.impl.dataflow.TransientNodeData;
import org.exoplatform.services.jcr.impl.dataflow.TransientPropertyData;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;
import org.exoplatform.services.jcr.impl.util.NodeDataReader;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL        .
 * @author Gennady Azarenkov
 * @version $Id: NodeTypeDataPersister.java 12841 2007-02-16 08:58:38Z peterit $
 */

public class NodeTypeDataPersister {
  
  public static Log log = ExoLogger.getLogger("jcr.NodeTypeDataPersister");
  
  private DataManager dataManager;
  
  private MutableItemDataChangesLog changesLog;
  
  private NodeData ntRoot;
  
  private ValueFactoryImpl valueFactory = null;
  
  private LocationFactory locationFactory;
  
  public NodeTypeDataPersister(DataManager dataManager, 
      ValueFactoryImpl valueFactory, LocationFactory locationFactory) {
    this.dataManager = dataManager;
    this.valueFactory = valueFactory;
    this.locationFactory = locationFactory;
    this.changesLog = new MutableItemDataChangesLog();
    try {
      this.ntRoot = (NodeData) dataManager.getItemData(Constants.JCR_NODETYPES_PATH);
    } catch (RepositoryException e) {
      log.warn("Nodetypes storage (/jcr:system/jcr:nodetypes node) is not initialized.");
    }
  }
  
  boolean isPersisted() {
    return ntRoot != null;
  }
   
  /**
   * Parse name in form of JCR(JSR-170) names conversion string. e.g. name_space:item_name, nt:base 
   * 
   * @param name
   * @return
   * @throws IllegalNameException
   */
  private InternalQName parseName(String name) throws RepositoryException {
    
    return locationFactory.parseJCRName(name).getInternalName();
  }
  
  void saveChanges() throws RepositoryException, InvalidItemStateException {
    dataManager.save(changesLog); //log.info(changesLog.dump())
    changesLog.clear();
  }  

  public synchronized void initNodetypesRoot(NodeData nsSystem, boolean addACL) {
    
    if (ntRoot == null) {
      long start = System.currentTimeMillis();
      
      TransientNodeData root = TransientNodeData.createNodeData(nsSystem,
          Constants.JCR_NODETYPES, Constants.NT_UNSTRUCTURED, Constants.NODETYPESROOT_UUID);
      
      TransientPropertyData primaryType = TransientPropertyData.createPropertyData(
          root, Constants.JCR_PRIMARYTYPE, PropertyType.NAME, false);
      primaryType.setValue(new TransientValueData(root.getPrimaryTypeName()));
      
      changesLog.add(ItemState.createAddedState(root))
                .add(ItemState.createAddedState(primaryType));
      
      if(addACL) {
        AccessControlList acl = new AccessControlList(); 
        root.setMixinTypeNames(new InternalQName[] {Constants.EXO_ACCESS_CONTROLLABLE} );
        // jcr:mixinTypes
        TransientPropertyData rootMixinTypes = TransientPropertyData
            .createPropertyData(root, Constants.JCR_MIXINTYPES, PropertyType.NAME, false);
        rootMixinTypes.setValue(new TransientValueData(Constants.EXO_ACCESS_CONTROLLABLE));

        TransientPropertyData exoOwner = TransientPropertyData.createPropertyData(
            root, Constants.EXO_OWNER, PropertyType.STRING, false);
        exoOwner.setValue(new TransientValueData(acl.getOwner()));
        TransientPropertyData exoPerms = TransientPropertyData.createPropertyData(
            root, Constants.EXO_PERMISSIONS, ExtendedPropertyType.PERMISSION, true);
        List<ValueData> perms = new ArrayList<ValueData>();
        for(int i=0; i<acl.getPermissionEntries().size(); i++) {
          AccessControlEntry entry = acl.getPermissionEntries().get(i);
          perms.add(new TransientValueData(entry));
        }
        exoPerms.setValues(perms);
        changesLog.add(ItemState.createAddedState(rootMixinTypes))
                  .add(ItemState.createAddedState(exoOwner))
                  .add(ItemState.createAddedState(exoPerms));
        changesLog.add(new ItemState(root, ItemState.MIXIN_CHANGED, false, null));
      }
      ntRoot = root;
      log.info("/jcr:system/jcr:nodetypes is created, creation time: "
          + (System.currentTimeMillis() - start) + " ms");
    } else {
      log.warn("/jcr:system/jcr:nodetypes already exists");
    }
  }
  
  public synchronized void initStorage(List<NodeType> nodetypes)
      throws PathNotFoundException, RepositoryException {

    if (!isPersisted()) {
      log.warn("Nodetypes storage (/jcr:system/jcr:nodetypes node) is not exists. Possible is not initialized (call initNodetypesRoot() before)");
      return;
    }
    long ntStart = System.currentTimeMillis();
    for (NodeType nt: nodetypes) {
      try {
        addNodeType(nt);
        
//        if (nt.getName().equals("nt:file")) {
//          for (PropertyDefinition pd: nt.getPropertyDefinitions()) {
//            if (pd.getName().equals("jcr:created")) {
//              log.info("nt:file-->jcr:created >>> " + OnParentVersionAction.nameFromValue(pd.getOnParentVersion()));
//            }
//          }
//        }
        log.info("Node type " + nt.getName() + " is initialized. ");
      } catch (ItemExistsException e) {
        log.warn("Node exists " + nt.getName() + ". Error: " + e.getMessage());
      }
    }
    saveChanges();
    log.info("Node types initialized. Time: "+ (System.currentTimeMillis() - ntStart) + " ms");

  }

  public boolean hasNodeTypeData(String nodeTypeName) throws RepositoryException {
    return hasNodeTypeData(parseName(nodeTypeName));
  }
  
  public boolean hasNodeTypeData(InternalQName nodeTypeName) throws RepositoryException {
    try {
      return getNodeTypesData(nodeTypeName).size()>0;
    } catch(PathNotFoundException e) {
      return false;
    }
  }
  
  private List<NodeDataReader> getNodeTypesData(InternalQName nodeTypeName) throws RepositoryException {
    
    NodeDataReader ntReader = new NodeDataReader(ntRoot, dataManager, valueFactory);
    ntReader.forNode(nodeTypeName);
    ntReader.read();
    
    ntReader.getNodes(nodeTypeName);
    
    return ntReader.getNodes(nodeTypeName);
  }
  
  public NodeData addNodeType(NodeType nodeType)
      throws PathNotFoundException, RepositoryException, ValueFormatException {

    if (!isPersisted()) {
      log.warn("Nodetypes storage (/jcr:system/jcr:nodetypes node) is not initialized.");
      return null;
    }
    
    NodeData ntNode = TransientNodeData.createNodeData(ntRoot, 
        parseName(nodeType.getName()), Constants.NT_NODETYPE);
    
    TransientPropertyData primaryType = TransientPropertyData.createPropertyData(
        ntNode, Constants.JCR_PRIMARYTYPE, PropertyType.NAME, false);
    primaryType.setValue(new TransientValueData(ntNode.getPrimaryTypeName()));

    TransientPropertyData name = TransientPropertyData.createPropertyData(ntNode, 
        Constants.JCR_NODETYPENAME, PropertyType.NAME, false); // jcr:nodeTypeName
    name.setValue(new TransientValueData(parseName(nodeType.getName())));

    TransientPropertyData isMixin = TransientPropertyData.createPropertyData(ntNode, 
        Constants.JCR_ISMIXIN, PropertyType.BOOLEAN, false); // jcr:isMixin
    isMixin.setValue(new TransientValueData(nodeType.isMixin()));

    TransientPropertyData hasOrderableChildNodes = TransientPropertyData.createPropertyData(ntNode, 
        Constants.JCR_HASORDERABLECHILDNODES, PropertyType.BOOLEAN, false); // jcr:hasOrderableChildNodes
    hasOrderableChildNodes.setValue(new TransientValueData(nodeType.hasOrderableChildNodes()));

    changesLog.add(ItemState.createAddedState(ntNode))
      .add(ItemState.createAddedState(primaryType))
      .add(ItemState.createAddedState(name))
      .add(ItemState.createAddedState(isMixin))
      .add(ItemState.createAddedState(hasOrderableChildNodes));
    
    //log.info("ntNode: " + ntNode.getPrimaryTypeName().getAsString() + " - " + ntNode.getQPath().getAsString());
    
    if (nodeType.getPrimaryItemName() != null) {
      TransientPropertyData primaryItemName = TransientPropertyData.createPropertyData(ntNode, 
          Constants.JCR_PRIMARYITEMNAME, PropertyType.NAME, false); // jcr:primaryItemName
      primaryItemName.setValue(new TransientValueData(parseName(nodeType.getPrimaryItemName())));
      changesLog.add(ItemState.createAddedState(primaryItemName));
    }
    List <ValueData> parents = new ArrayList<ValueData>(); 
    for (int i = 0; i < nodeType.getDeclaredSupertypes().length; i++) {
      parents.add(new TransientValueData(parseName(nodeType.getDeclaredSupertypes()[i].getName())));
    }
    if (parents.size() != 0) {
      TransientPropertyData supertypes = TransientPropertyData.createPropertyData(
          ntNode, Constants.JCR_SUPERTYPES, PropertyType.NAME, true); // jcr:supertypes
      supertypes.setValues(parents);
      changesLog.add(ItemState.createAddedState(supertypes));
    }
    
    if (nodeType.getDeclaredPropertyDefinitions().length > 0) {
      //String pdQName = Constants.JCR_PROPERTYDEFINITION.getAsString(); // SameNameSiblings true
      for (int i = 0; i < nodeType.getDeclaredPropertyDefinitions().length; i++) {
        //NodeData childProps = TransientNodeData.createNodeData(ntNode,
        //    InternalQName.parse(pdQName + ":" + (i + 1)), Constants.NT_PROPERTYDEFINITION);
        NodeData childProps = TransientNodeData.createNodeData(ntNode,
            Constants.JCR_PROPERTYDEFINITION, Constants.NT_PROPERTYDEFINITION, i + 1);
        
        TransientPropertyData cpPrimaryType = TransientPropertyData.createPropertyData(
            childProps, Constants.JCR_PRIMARYTYPE, PropertyType.NAME, false);
        cpPrimaryType.setValue(new TransientValueData(childProps.getPrimaryTypeName()));
        
        changesLog.add(ItemState.createAddedState(childProps))
                  .add(ItemState.createAddedState(cpPrimaryType));

        initPropertyDefProps(childProps, nodeType.getDeclaredPropertyDefinitions()[i]);
      }
    }

    if (nodeType.getDeclaredChildNodeDefinitions().length > 0) { 
      //String cdQName = Constants.JCR_CHILDNODEDEFINITION.getAsString(); // SameNameSiblings true
      for (int i = 0; i < nodeType.getDeclaredChildNodeDefinitions().length; i++) {
        //NodeData childNodes = TransientNodeData.createNodeData(ntNode,
        //    InternalQName.parse(cdQName + ":" + (i + 1)), Constants.NT_CHILDNODEDEFINITION, i);
        NodeData childNodes = TransientNodeData.createNodeData(ntNode,
            Constants.JCR_CHILDNODEDEFINITION, Constants.NT_CHILDNODEDEFINITION, i + 1);
        
        TransientPropertyData cnPrimaryType = TransientPropertyData.createPropertyData(
            childNodes, Constants.JCR_PRIMARYTYPE, PropertyType.NAME, false);
        cnPrimaryType.setValue(new TransientValueData(childNodes.getPrimaryTypeName()));
        
        changesLog.add(ItemState.createAddedState(childNodes))
                  .add(ItemState.createAddedState(cnPrimaryType));
        //log.info("  childNodes: " + childNodes.getPrimaryTypeName().getAsString() + " - " + childNodes.getQPath().getAsString());
        initNodeDefProps(childNodes, nodeType.getDeclaredChildNodeDefinitions()[i]);
      }
    }
    return ntNode;
  }

  private void initPropertyDefProps(NodeData parent, PropertyDefinition def)
      throws ValueFormatException, RepositoryException {

    if (def.getName() != null) { //Mandatory false
      TransientPropertyData name = TransientPropertyData.createPropertyData(
          parent, Constants.JCR_NAME, PropertyType.NAME, false);
      name.setValue(new TransientValueData(parseName(def.getName())));
      changesLog.add(ItemState.createAddedState(name));
    }
    
    TransientPropertyData autoCreated = TransientPropertyData.createPropertyData(
        parent, Constants.JCR_AUTOCREATED, PropertyType.BOOLEAN, false);
    autoCreated.setValue(new TransientValueData(def.isAutoCreated()));
    
    TransientPropertyData isMandatory = TransientPropertyData.createPropertyData(
        parent, Constants.JCR_MANDATORY, PropertyType.BOOLEAN, false);
    isMandatory.setValue(new TransientValueData(def.isMandatory()));
    
    TransientPropertyData onParentVersion = TransientPropertyData.createPropertyData(
        parent, Constants.JCR_ONPARENTVERSION, PropertyType.STRING, false);
    onParentVersion.setValue(new TransientValueData(
        OnParentVersionAction.nameFromValue(def.getOnParentVersion())));
    
    TransientPropertyData isProtected = TransientPropertyData.createPropertyData(
        parent, Constants.JCR_PROTECTED, PropertyType.BOOLEAN, false);
    isProtected.setValue(new TransientValueData(def.isProtected()));
    
    TransientPropertyData requiredType = TransientPropertyData.createPropertyData(
        parent, Constants.JCR_REQUIREDTYPE, PropertyType.STRING, false);
    requiredType.setValue(new TransientValueData(
        ExtendedPropertyType.nameFromValue(def.getRequiredType())));
    
    TransientPropertyData isMultiple = TransientPropertyData.createPropertyData(
        parent, Constants.JCR_MULTIPLE, PropertyType.BOOLEAN, false);
    isMultiple.setValue(new TransientValueData(def.isMultiple()));
    
    changesLog.add(ItemState.createAddedState(autoCreated))
      .add(ItemState.createAddedState(isMandatory))
      .add(ItemState.createAddedState(onParentVersion))
      .add(ItemState.createAddedState(isProtected))
      .add(ItemState.createAddedState(requiredType))
      .add(ItemState.createAddedState(isMultiple));
    
    if (def.getValueConstraints() != null && def.getValueConstraints().length != 0) {
      List <ValueData> valueConstraintsValues = new ArrayList<ValueData>(); 
      for (int i = 0; i < def.getValueConstraints().length; i++) {
        valueConstraintsValues.add(new TransientValueData(def.getValueConstraints()[i]));
      }
      TransientPropertyData valueConstraints = TransientPropertyData.createPropertyData(
          parent, Constants.JCR_VALUECONSTRAINTS, PropertyType.STRING, true);
      valueConstraints.setValues(valueConstraintsValues);
      changesLog.add(ItemState.createAddedState(valueConstraints));
    }
    
    if (def.getDefaultValues() != null && def.getDefaultValues().length != 0) {
      List <ValueData> defaultValuesValues = new ArrayList<ValueData>(); 
      for (int i = 0; i < def.getDefaultValues().length; i++) {
        // [PN] 27.07.06
        if (def.getDefaultValues()[i] != null)
          defaultValuesValues.add(new TransientValueData(def.getDefaultValues()[i].getString()));
      }
      TransientPropertyData defaultValues = TransientPropertyData.createPropertyData(
          parent, Constants.JCR_DEFAULTVALUES, PropertyType.STRING, true);
      defaultValues.setValues(defaultValuesValues);
      changesLog.add(ItemState.createAddedState(defaultValues));
    }
  }

  private void initNodeDefProps(NodeData parent, NodeDefinition def)
      throws ValueFormatException, RepositoryException {

    if (def.getName() != null) { //Mandatory false
      TransientPropertyData name = TransientPropertyData.createPropertyData(
          parent, Constants.JCR_NAME, PropertyType.NAME, false);
      name.setValue(new TransientValueData(parseName(def.getName())));
      changesLog.add(ItemState.createAddedState(name));
    }
    
    TransientPropertyData autoCreated = TransientPropertyData.createPropertyData(
        parent, Constants.JCR_AUTOCREATED, PropertyType.BOOLEAN, false);
    autoCreated.setValue(new TransientValueData(def.isAutoCreated()));
    
    TransientPropertyData isMandatory = TransientPropertyData.createPropertyData(
        parent, Constants.JCR_MANDATORY, PropertyType.BOOLEAN, false);
    isMandatory.setValue(new TransientValueData(def.isMandatory()));
    
    TransientPropertyData onParentVersion = TransientPropertyData.createPropertyData(
        parent, Constants.JCR_ONPARENTVERSION, PropertyType.STRING, false);
    onParentVersion.setValue(new TransientValueData(
        OnParentVersionAction.nameFromValue(def.getOnParentVersion())));
    
    TransientPropertyData isProtected = TransientPropertyData.createPropertyData(
        parent, Constants.JCR_PROTECTED, PropertyType.BOOLEAN, false);
    isProtected.setValue(new TransientValueData(def.isProtected()));
    
    TransientPropertyData sameNameSiblings = TransientPropertyData.createPropertyData(
        parent, Constants.JCR_SAMENAMESIBLINGS, PropertyType.BOOLEAN, false);
    sameNameSiblings.setValue(new TransientValueData(def.allowsSameNameSiblings()));
    
    if (def.getDefaultPrimaryType() != null) { //Mandatory false
      TransientPropertyData defaultPrimaryType = TransientPropertyData.createPropertyData(
          parent, Constants.JCR_DEFAULTPRIMNARYTYPE, PropertyType.NAME, false);
      defaultPrimaryType.setValue(new TransientValueData(parseName(def.getDefaultPrimaryType().getName())));
      changesLog.add(ItemState.createAddedState(defaultPrimaryType));
    }
    
    changesLog.add(ItemState.createAddedState(autoCreated))
      .add(ItemState.createAddedState(isMandatory))
      .add(ItemState.createAddedState(onParentVersion))
      .add(ItemState.createAddedState(isProtected))
      .add(ItemState.createAddedState(sameNameSiblings));    
    
    if (def.getRequiredPrimaryTypes() != null && def.getRequiredPrimaryTypes().length != 0) {
      List <ValueData> requiredPrimaryTypesValues = new ArrayList<ValueData>(); 
      for (int i = 0; i < def.getRequiredPrimaryTypes().length; i++) {
        requiredPrimaryTypesValues.add(new TransientValueData(parseName(def.getRequiredPrimaryTypes()[i].getName())));
      }
      TransientPropertyData requiredPrimaryTypes = TransientPropertyData.createPropertyData(
          parent, Constants.JCR_REQUIREDPRIMARYTYPES, PropertyType.NAME, true);
      requiredPrimaryTypes.setValues(requiredPrimaryTypesValues);
      changesLog.add(ItemState.createAddedState(requiredPrimaryTypes));
    }  
  }

  private NodeType findType(String nodeTypeName, List<NodeType> ntList) {
    for (NodeType regNt: ntList) {
      if (regNt.getName().equals(nodeTypeName)) {
        return regNt;
      }
    }
    return null;
  }
    
  public List<NodeType> loadNodetypes(List<NodeType> registeredNodeTypes, NodeTypeManagerImpl ntManager) throws PathNotFoundException, RepositoryException {
    
    if (!isPersisted())
      ntRoot = (NodeData) dataManager.getItemData(Constants.JCR_NODETYPES_PATH);
    
    if (!isPersisted()) {
      log.warn("Nodetypes storage (/jcr:system/jcr:nodetypes node) is not initialized. No nodetypes loaded.");
      return new ArrayList<NodeType>();
    }
    
    List<NodeType> ntList = new ArrayList<NodeType>();
    List<NodeType> loadedList = new ArrayList<NodeType>();
    ntList.addAll(registeredNodeTypes);
    
    boolean nextCycle = false;
    List<NodeType> registeringTypes = new ArrayList<NodeType>();
    
    int registerCyclesCount = 1;
    do {
      
      long cycleStart = System.currentTimeMillis();
      log.info(">>> Node types registration cycle " + registerCyclesCount + " started");
      
      NodeDataReader ntReader = new NodeDataReader(ntRoot, dataManager, valueFactory);
      ntReader.forNodesByType(Constants.NT_NODETYPE); // for nt:nodeType
      ntReader.read();
      
      nextNodeType: for (NodeDataReader ntr: ntReader.getNodesByType(Constants.NT_NODETYPE)) {
        
        long ntStart = System.currentTimeMillis();
        
        ntr.forProperty(Constants.JCR_NODETYPENAME, PropertyType.NAME);
        ntr.read();
        String ntName = ntr.getPropertyValue(Constants.JCR_NODETYPENAME).getString();
        NodeType existedNodeType = findType(ntName, ntList);
        if (existedNodeType != null) {
          if (log.isDebugEnabled())
            log.debug("Already reagistered " + ntName);
          continue nextNodeType; // already registered node type with this name
        }
        if (log.isDebugEnabled())
          log.debug("Reagistering from storage " + ntName + " " + (System.currentTimeMillis() - ntStart));
        
        //ntr.forProperty(Constants.JCR_NODETYPENAME, PropertyType.NAME)
        ntr.forProperty(Constants.JCR_PRIMARYTYPE, PropertyType.NAME)
           .forProperty(Constants.JCR_ISMIXIN, PropertyType.BOOLEAN)
           .forProperty(Constants.JCR_HASORDERABLECHILDNODES, PropertyType.BOOLEAN)
           .forProperty(Constants.JCR_PRIMARYITEMNAME, PropertyType.NAME)
           .forProperty(Constants.JCR_SUPERTYPES, PropertyType.NAME);
        ntr.forNodesByType(Constants.NT_PROPERTYDEFINITION)
           .forNodesByType(Constants.NT_CHILDNODEDEFINITION);
        ntr.read();
        if (log.isDebugEnabled())
          log.debug("Node type readed " + ntName + " " + (System.currentTimeMillis() - ntStart));
        
        NodeTypeImpl type = new NodeTypeImpl(ntManager);
        //type.setName(parseName(ntr.getPropertyValue(Constants.JCR_NODETYPENAME).getString()).getAsString());
        type.setName(ntr.getPropertyValue(Constants.JCR_NODETYPENAME).getString());
        type.setMixin(ntr.getPropertyValue(Constants.JCR_ISMIXIN).getBoolean());
        type.setOrderableChild(ntr.getPropertyValue(Constants.JCR_HASORDERABLECHILDNODES).getBoolean());
        try {
          type.setPrimaryItemName(ntr.getPropertyValue(Constants.JCR_PRIMARYITEMNAME).getString());
        } catch(PathNotFoundException e) { // Mandatory false
        }
        
        if (!registeringTypes.contains(type))
          registeringTypes.add(type);
        
        // -------- Super types --------
        try {
          List<Value> dst = ntr.getPropertyValues(Constants.JCR_SUPERTYPES);
          NodeType[] declaredSupertypes = new NodeType[dst.size()];
          for (int i = 0; i < dst.size(); i++) {
            String superTypeName = dst.get(i).getString();
            declaredSupertypes[i] = findType(superTypeName, ntList);
            if (declaredSupertypes[i] == null) {
              if (nextCycle && findType(superTypeName, registeringTypes) == null)
                // here is this type and we try register 
                throw new ConstraintViolationException("Supertype " + superTypeName 
                    + " is not registered in repository (but need to be registered before nodetype " + type.getName() 
                    + "). Node type resistration aborted.");
              log.info("Supertype " + superTypeName 
                  + " is not registered. " + type.getName() 
                  + " node type will be registered in a next cycle.");
              continue nextNodeType;
            }
          }
          type.setDeclaredSupertypes(declaredSupertypes);
        } catch(PathNotFoundException e) {
        }
  
        // -------- Property definitions --------
        if (log.isDebugEnabled())
          log.debug("Property definitions for " + ntName + " " + (System.currentTimeMillis() - ntStart));
        try {
          List<NodeDataReader> pdNodes = ntr.getNodesByType(Constants.NT_PROPERTYDEFINITION);
          PropertyDefinition[] declaredPropertyDefs = new PropertyDefinition[pdNodes.size()];
          for (int pdi=0; pdi<pdNodes.size(); pdi++) {
            NodeDataReader pdr = pdNodes.get(pdi);
            
            pdr.forProperty(Constants.JCR_NAME, PropertyType.NAME) // jcr:name
               .forProperty(Constants.JCR_AUTOCREATED, PropertyType.BOOLEAN) // jcr:autoCreated
               .forProperty(Constants.JCR_MANDATORY, PropertyType.BOOLEAN) // jcr:mandatory
               .forProperty(Constants.JCR_PROTECTED, PropertyType.BOOLEAN) // jcr:protected
               .forProperty(Constants.JCR_MULTIPLE, PropertyType.BOOLEAN) // jcr:multiple
               .forProperty(Constants.JCR_ONPARENTVERSION, PropertyType.STRING) // jcr:onParentVersion
               //.forProperty(Constants.JCR_PRIMARYTYPE, PropertyType.NAME) // jcr:primaryType
               .forProperty(Constants.JCR_REQUIREDTYPE, PropertyType.STRING) // jcr:requiredType
               .forProperty(Constants.JCR_VALUECONSTRAINTS, PropertyType.STRING) // jcr:valueConstraints
               .forProperty(Constants.JCR_DEFAULTVALUES, PropertyType.STRING); // jcr:defaultValues        
            pdr.read();
            
            PropertyDefinitionImpl pDef = new PropertyDefinitionImpl();
            try {
              NameValue nameValue = (NameValue) pdr.getPropertyValue(Constants.JCR_NAME);
              pDef.setName(nameValue.getString());
              pDef.setQName(nameValue.getQName());
            } catch(PathNotFoundException e) {
            }
            try {
              pDef.setAutoCreate(pdr.getPropertyValue(Constants.JCR_AUTOCREATED).getBoolean());
              pDef.setMandatory(pdr.getPropertyValue(Constants.JCR_MANDATORY).getBoolean());
              pDef.setReadOnly(pdr.getPropertyValue(Constants.JCR_PROTECTED).getBoolean());
              pDef.setMultiple(pdr.getPropertyValue(Constants.JCR_MULTIPLE).getBoolean());
              pDef.setOnVersion(
                  OnParentVersionAction.valueFromName(
                      pdr.getPropertyValue(Constants.JCR_ONPARENTVERSION).getString()));
              
              pDef.setDeclaringNodeType(type);
              pDef.setRequiredType(
                  ExtendedPropertyType.valueFromName(
                      pdr.getPropertyValue(Constants.JCR_REQUIREDTYPE).getString()));
            } catch (PathNotFoundException e) {
              throw new ConstraintViolationException("Mandatory property did not set."
                  + " PropertyDefinition: " + pDef.getName() + ". Type: " + type.getName() + ". Error: "
                  + e.getMessage() + ". Node type resistration aborted.");
            }
            
            try {
              List<Value> valueConstraintValues = pdr.getPropertyValues(Constants.JCR_VALUECONSTRAINTS);
              String[] valueConstraints = new String[valueConstraintValues.size()];
              for (int j = 0; j < valueConstraintValues.size(); j++) {
                if (valueConstraintValues.get(j) != null)
                  valueConstraints[j] = valueConstraintValues.get(j).getString();
                else
                  valueConstraints[j] = null;
              }
              pDef.setValueConstraints(valueConstraints);
            } catch (PathNotFoundException e) { // Mandatory false
            }
            
            try {
              List<Value> dvl = pdr.getPropertyValues(Constants.JCR_DEFAULTVALUES);
              Value[] defaultValues = new Value[dvl.size()];
              for (int i=0; i<dvl.size(); i++) {
                defaultValues[i] = dvl.get(i);
              }
              pDef.setDefaultValues(defaultValues);
            } catch (PathNotFoundException e) { // Mandatory false
            }
            
            if (log.isDebugEnabled())
              log.debug("Property definitions readed " + pDef.getName() + " " + (System.currentTimeMillis() - ntStart));
            
            declaredPropertyDefs[pdi] = pDef;
          }
          type.setDeclaredPropertyDefs(declaredPropertyDefs);
        } catch(PathNotFoundException e) { // Mandatory false
        }
  
        // --------- Child nodes definitions ----------
        if (log.isDebugEnabled())
          log.debug("Child nodes definitions for " + ntName + " " + (System.currentTimeMillis() - ntStart));
        try {
          List<NodeDataReader> cdNodes = ntr.getNodesByType(Constants.NT_CHILDNODEDEFINITION);
          NodeDefinition[] declaredChildNodesDefs = new NodeDefinition[cdNodes.size()];
          for (int cdi=0; cdi<cdNodes.size(); cdi++) {
            NodeDataReader cdr = cdNodes.get(cdi);
            
            cdr.forProperty(Constants.JCR_NAME, PropertyType.NAME) // jcr:name
               .forProperty(Constants.JCR_REQUIREDPRIMARYTYPES, PropertyType.NAME) // jcr:requiredPrimaryTypes
               .forProperty(Constants.JCR_AUTOCREATED, PropertyType.BOOLEAN) // jcr:autoCreated
               .forProperty(Constants.JCR_MANDATORY, PropertyType.BOOLEAN) // jcr:mandatory
               .forProperty(Constants.JCR_PROTECTED, PropertyType.BOOLEAN) // jcr:protected
               .forProperty(Constants.JCR_ONPARENTVERSION, PropertyType.STRING) // jcr:onParentVersion
               .forProperty(Constants.JCR_SAMENAMESIBLINGS, PropertyType.STRING) // jcr:sameNameSiblings
               //.forProperty(Constants.JCR_PRIMARYTYPE, PropertyType.STRING) // jcr:primaryType
               .forProperty(Constants.JCR_DEFAULTPRIMNARYTYPE, PropertyType.NAME); // jcr:defaultPrimaryType
            cdr.read();
            
            NodeDefinitionImpl nDef = new NodeDefinitionImpl();
            try {
              NameValue nameValue = (NameValue) cdr.getPropertyValue(Constants.JCR_NAME);
              nDef.setName(nameValue.getString());
              nDef.setQName(nameValue.getQName());
            } catch (PathNotFoundException e) { // Mandatory false
            }
            try {
              String defaultNodeTypeName = cdr.getPropertyValue(Constants.JCR_DEFAULTPRIMNARYTYPE).getString();
              NodeType defaultNodeType = findType(defaultNodeTypeName, ntList);
              if (defaultNodeType != null)
                nDef.setDefaultNodeType(defaultNodeType);
              else if (defaultNodeType == null && defaultNodeTypeName.equals(type.getName())) 
                nDef.setDefaultNodeType(type);
            } catch (PathNotFoundException e) { // Mandatory false
            }
            try {
              List<Value> requiredNodeTypesValues = cdr.getPropertyValues(Constants.JCR_REQUIREDPRIMARYTYPES);
              NodeType[] requiredNodeTypes = new NodeType[requiredNodeTypesValues.size()];
              for (int j = 0; j < requiredNodeTypesValues.size(); j++) {
                if (requiredNodeTypesValues.get(j) != null) {
                  String requiredNodeTypeName = requiredNodeTypesValues.get(j).getString();
                  NodeType requiredNodeType = findType(requiredNodeTypeName, ntList);
                  if (requiredNodeType != null) {
                    requiredNodeTypes[j] = requiredNodeType;
                  } else {
                    // [PN] 10.01.07 Fix search of requiredNodeTypeName by String name
                    //if (nextCycle && !registeringTypes.contains(requiredNodeTypeName))
                    if (nextCycle && findType(requiredNodeTypeName, registeringTypes) == null)
                      throw new ConstraintViolationException(
                          "Required node type of NodeDefinition "
                          + nDef.getName() + " is not registered in repository. Required node type " 
                          + requiredNodeTypeName + " must be registered before " + type.getName() 
                          + ". Node type resistration aborted.");
                    log.info("Required node type of NodeDefinition "
                        + nDef.getName() + " is not registered." + type.getName() 
                        + " node type will be registered in a next cycle.");
                    continue nextNodeType;
                  }
                } else {
                  throw new ConstraintViolationException("Required node type is null."
                      + " Type: " + type.getName() + ". NodeDefinition: " + nDef.getName()
                      + ". Node type resistration aborted.");
                }
              }
              nDef.setRequiredNodeTypes(requiredNodeTypes);
              nDef.setDeclaringNodeType(type);
              nDef.setAutoCreate(cdr.getPropertyValue(Constants.JCR_AUTOCREATED).getBoolean());
              nDef.setMandatory(cdr.getPropertyValue(Constants.JCR_MANDATORY).getBoolean());
              nDef.setReadOnly(cdr.getPropertyValue(Constants.JCR_PROTECTED).getBoolean());
              nDef.setMultiple(cdr.getPropertyValue(Constants.JCR_SAMENAMESIBLINGS).getBoolean());
              nDef.setOnVersion(
                  OnParentVersionAction.valueFromName(
                      cdr.getPropertyValue(Constants.JCR_ONPARENTVERSION).getString()));
            } catch (PathNotFoundException e) {
              throw new ConstraintViolationException("Mandatory property did not set."
                  + " NodeDefinition: " + nDef.getName() + ". Type: " + type.getName() + ". Error: "
                  + e.getMessage() + ". Node type resistration aborted.", e);
            }
            
            if (log.isDebugEnabled())
              log.debug("Child nodes definitions readed " + nDef.getName() + " " + (System.currentTimeMillis() - ntStart));
            
            declaredChildNodesDefs[cdi] = nDef;
          }
          type.setDeclaredNodeDefs(declaredChildNodesDefs);
        } catch(PathNotFoundException e) { // Mandatory false
        }
        
        // -------- NodeType done --------
        ntList.add(type);
        loadedList.add(type);
        //unregisterdTypesNames.remove(type.getName());
        log.info("NodeType " + type.getName() + " loaded. " + (System.currentTimeMillis() - ntStart) + " ms");
      }
      nextCycle = true;
      log.info("<<< Node types registration cycle " + registerCyclesCount + " finished. "
          + (System.currentTimeMillis() - cycleStart) + " ms");
      registerCyclesCount++; // for owerflow limitation
      if (registerCyclesCount >= 1000)
        throw new RepositoryException("Maximum cycles count of NodeType registrations reached, 1000. Registration breaked."); 
    } while (registeringTypes.size() > loadedList.size());
    return loadedList;
  }  
 
}
